/*
 * Author: Randy Merkel
 *
 *------------------------------------------------------------------*/

#include <windows.h>
#include <stdio.h>
#include <math.h>
#include <iostream>
#include <time.h>
#include <assert.h>
#include <functional>
#include <algorithm>
#include <list>
#include <set>
#include "Model.h"
#include "MyVector.h"
#include "GraphNode.h"
#include  <glut.h>

using namespace std;

#define ZINCREMENT 0.1f
#define INCREMENT .1f
#define PI 3.14159f
#define BIG_NUM 10000.0f

// some function prototypes
void generateEdges();
void display(void);
void normalize(float[3]);
void normCrossProd(float[3], float[3], float[3]);
void drawMesh();
void drawFace(Face *, int, int);
void drawEdge(const Edge &, bool);
void drawRootNormal();

static bool debugMode = true;

// initial viewer position
static GLdouble viewer[] = {0.0, 5.0, 100.0};

const double MAX_RAD = 10;
const double MIN_RAD = 0.2;
const double MAX_PHI = PI / 3;
const double MIN_PHI = -PI / 3;

const double ROOT_NORM_SCALE = 200;

int GW;
int GH;

const double DRAW_THRESH = -10.0;

float phi, cPhi;
float theta, cTheta;
float radius, cRadius;

static int rootI;
static MyVector rootV;
static int rootFace = -1;

MyVector cCamera;
float cLight[4] = {2.0, 2.0, 2.0, 1.0};

Model *gModel = NULL;
vector<int> *vertexXFace = NULL;

static int jumpEdgeCount = 0;

static size_t facesVisited = 0;

vector<GraphNode> *gVGraph = NULL;
vector<GraphNode> *gFGraph = NULL;
vector<Edge> boundaryEdge;
bool *faceVisited = NULL;
Edge badEdge(-1,-1);

bool *multVisited = NULL;

bool drawTrees = false;

struct materialStruct {
  float ambient[4];
  float diffuse[4];
  float specular[4];
  float shininess[1];
};

materialStruct Jade = {
  {0.135, 0.2225, 0.1575, 0.95},
  {0.54, 0.89, 0.63, 0.95},
  {0.316, 0.316, 0.316, 0.95},
  {27.89}
};

materialStruct RedFlat = {
  {0.3, 0.0, 0.0, 1.0},
  {0.9, 0.0, 0.0, 1.0},
  {0.0, 0.0, 0.0, 1.0},
  {0.0}
};
materialStruct GreenShiny = {
  {0.0, 0.3, 0.0, 1.0},
  {0.0, 0.9, 0.0, 1.0},
  {0.2, 1.0, 0.2, 1.0},
  {0.8}
};
materialStruct Pearl = {
  {.25, .20725, .20725, 1.0},
  {1, .829, .829, 1.0},
  {.296648, .296648, .296648, 1.0},
  {.088}
};
materialStruct Obsidian = {
  {.05375, .05, .06625, 1.0},
  {.18275, .17, .22525, 1.0},
  {.332741, .328634, .346435, 1.0},
  {.3}
};
materialStruct Chrome = {
  {.25, .25, .25, 1.0},
  {.4, .4, .4, 1.0},
  {.774597, .774597, .774597, 1.0},
  {.6}
};
materialStruct Gold = {
  {.24725, .1995, .0745, 1.0},
  {.75164, .60648, .22648, 1.0},
  {.628281, .555802, .366065, 1.0},
  {.4}
};

materialStruct WhiteFlat = {
  {.5, .5, .5, 1.0},
  {1, 1, 1, 1.0},
  {0, 0, 0, 1.0},
  {0}
};

enum {kTree, kModel, kGraph};

static int drawMode = kModel;

static const int NUM_MATERIALS = 8;
materialStruct materials[NUM_MATERIALS] = {Jade, GreenShiny, RedFlat, Pearl, Obsidian, Chrome, Gold, WhiteFlat};

//light color (ambiant, diffuse and specular)
GLfloat light_amb[4] = {0.8, 0.8, 0.8, 1.0};
GLfloat light_diff[4] = {0.8, 0.8, 0.8, 1.0};
GLfloat light_spec[4] = {0.6, 0.6, 0.6, 1.0};

void addEdge(vector<vector<Edge> > *edges, int v1, int v2, int face) {
   Edge e(v1, v2);
   vector<Edge>::iterator eitr;
   MyVector midPnt;
   bool found = false;

   for (eitr = ((*edges)[e.v1 - 1]).begin(); eitr != ((*edges)[e.v1 - 1]).end(); eitr++) {
	   if ((*eitr).v2 == e.v2){
		  (*eitr).t2 = face;
		  found = true;
		  break;
	   }
   }

   if (!found) {
      e.t1 = face;
      (*edges)[e.v1 - 1].push_back(e);
   }
}

void generateEdges(Model *model){
   vector<Face *>::iterator fitr;
   Edge e1;
   int face;

   vertexXFace = new vector<int>[model->mVertices.size()];

   for (fitr = model->mFaces.begin(), face = 0; fitr != model->mFaces.end(); fitr++, face++) {
      addEdge(model->mEdges, (*fitr)->v1, (*fitr)->v2, face);
      addEdge(model->mEdges, (*fitr)->v2, (*fitr)->v3, face);
      addEdge(model->mEdges, (*fitr)->v3, (*fitr)->v1, face);

      vertexXFace[(*fitr)->v1 - 1].push_back(face);
      vertexXFace[(*fitr)->v2 - 1].push_back(face);
      vertexXFace[(*fitr)->v3 - 1].push_back(face);
   }
}

void mergeFGraph(Graph *g1, Graph *g2, vector<GraphNode> *graph) {
   vector<int>::const_iterator itr;

   for (itr = g2->getNodes().begin(); itr != g2->getNodes().end(); itr++) {
      g1->addNode(*itr);
      (*graph)[*itr].setGraph(g1);
   }

   delete g2;
}

void mergeFaces(const Edge &e, vector<GraphNode> *fGraph) {
   Graph *f1Graph = (*fGraph)[e.t1].getGraph();
   Graph *f2Graph = (*fGraph)[e.t2].getGraph();

   if (e.t1 != -1 && e.t2 != -1) {
      if (f1Graph) {
	      if (f1Graph != f2Graph) {
            if (f2Graph) {
               if (f1Graph->getSize() > f2Graph->getSize()) {
                  mergeFGraph(f1Graph, f2Graph, fGraph);
               } else {
                  mergeFGraph(f2Graph, f1Graph, fGraph);
               }
            } else {
               f1Graph->addNode(e.t2);
               (*fGraph)[e.t2].setGraph(f1Graph);
               facesVisited++;
            }

            (*fGraph)[e.t1].addSibling(e.t2);
            (*fGraph)[e.t2].addSibling(e.t1);
         // create cycle
         } else if (!(*fGraph)[e.t1].contains(e.t2)) {
            (*fGraph)[e.t1].addSibling(e.t2);
            (*fGraph)[e.t2].addSibling(e.t1);
         }
      } else if (f2Graph) {
         f2Graph->addNode(e.t1);
         (*fGraph)[e.t1].setGraph(f2Graph);

         (*fGraph)[e.t1].addSibling(e.t2);
         (*fGraph)[e.t2].addSibling(e.t1);
         facesVisited++;
      } else {
         f1Graph = new Graph();
         f1Graph->addNode(e.t1);
         f1Graph->addNode(e.t2);
         (*fGraph)[e.t1].setGraph(f1Graph);
         (*fGraph)[e.t2].setGraph(f1Graph);

         (*fGraph)[e.t1].addSibling(e.t2);
         (*fGraph)[e.t2].addSibling(e.t1);
         facesVisited++;
         facesVisited++;
      }
   }
}

const Edge &getEdge(int v1, int v2) {
   Edge e(v1, v2);
   vector<Edge>::iterator eitr;

   for (eitr = (*gModel->mEdges)[e.v1 - 1].begin(); eitr != (*gModel->mEdges)[e.v1 - 1].end(); eitr++) {
      if ((*eitr).v2 == e.v2) {
         return *eitr;
      }
   }
   cout << "Edge not found: " << v1 << ", " << v2 << endl;
   assert(false);

   return (*new Edge(1,2));
}

void mergeVGraph(Graph *g1, Graph *g2, vector<GraphNode> *graph) {
   vector<int>::const_iterator itr;

   for (itr = g2->getNodes().begin(); itr != g2->getNodes().end(); itr++) {
      g1->addNode(*itr);
      (*graph)[*itr - 1].setGraph(g1);
   }

   delete g2;
}

void mergeVertices(int v1, int v2, vector<GraphNode> *vGraph) {
   Graph *v1Graph = (*vGraph)[v1 - 1].getGraph();
   Graph *v2Graph = (*vGraph)[v2 - 1].getGraph();

   (*vGraph)[v1 - 1].addSibling(v2);
   (*vGraph)[v2 - 1].addSibling(v1);

   if (v1Graph) {
      if (v2Graph) {
         if (v1Graph->getSize() > v2Graph->getSize()) {
            mergeVGraph(v1Graph, v2Graph, vGraph);
         } else {
            mergeVGraph(v2Graph, v1Graph, vGraph);
         }
      } else {
         v1Graph->addNode(v2);
         (*vGraph)[v2 - 1].setGraph(v1Graph);
      }
   } else if (v2Graph) {
      v2Graph->addNode(v1);
      (*vGraph)[v1 - 1].setGraph(v2Graph);
   } else {
      v1Graph = new Graph();
      v1Graph->addNode(v1);
      v1Graph->addNode(v2);
      (*vGraph)[v1 - 1].setGraph(v1Graph);
      (*vGraph)[v2 - 1].setGraph(v1Graph);
   }
}

void processEdge(int v1, int v2, int vertex, vector<GraphNode> *vGraph, vector<GraphNode> *fGraph, set<int> *boundary) {
   Graph *v1Graph = (*vGraph)[v1 - 1].getGraph();
   Graph *v2Graph = (*vGraph)[v2 - 1].getGraph();
   Graph *vertexGraph = (*vGraph)[vertex - 1].getGraph();
   Edge edge;

   // if edge does not include vertex
   if (v1 != vertex && v2 != vertex && vertexGraph != v1Graph && vertexGraph != v2Graph) {
      boundary->insert(v1);
      boundary->insert(v2);
      // if vertices on same graph and not connected in vertex spanning tree
      if (v1Graph && v1Graph == v2Graph && !(*vGraph)[v1 - 1].contains(v2)) {
         //store vertices as leaves of boundary
         //merge faces
         edge = getEdge(v1, v2);
         mergeFaces(edge, fGraph);
         boundaryEdge.push_back(edge);
         //facesVisited++;
      // else
      } else if ((!v1Graph && !v2Graph) || v1Graph != v2Graph) {
         // merge vertices
         mergeVertices(v1, v2, vGraph);
      }
   // else if vertices not in same graph
   } else if (v1Graph != v2Graph || !(*vGraph)[v1 - 1].contains(v2) ||!v1Graph) {
      if (v1Graph != vertexGraph) boundary->insert(v1);
      if (v2Graph != vertexGraph) boundary->insert(v2);
      // merge triangles
      mergeFaces(getEdge(v1, v2), fGraph);
      //facesVisited++;
   // else
   } else {
      // do nothing
   }
}

void connectBoundaries(vector<GraphNode> *vGraph, vector<GraphNode> *fGraph, const Edge &edge) {
   Graph *graph;

   if (edge.t1 != -1 && edge.t2 != -1) {
      // disconnect the faces
      (*fGraph)[edge.t1].removeSibling(edge.t2);
      (*fGraph)[edge.t2].removeSibling(edge.t1);
      // connect the vertices
      (*vGraph)[edge.v1 - 1].addSibling(edge.v2);
      (*vGraph)[edge.v2 - 1].addSibling(edge.v1);
      if ((*vGraph)[edge.v1 - 1].getGraph()) {
         if ((*vGraph)[edge.v2 - 1].getGraph()) {
            if ((*vGraph)[edge.v1 - 1].getGraph() != (*vGraph)[edge.v2 - 1].getGraph()) {
               mergeVGraph((*vGraph)[edge.v1 - 1].getGraph(), (*vGraph)[edge.v2 - 1].getGraph(), vGraph);
            }
         } else {
            (*vGraph)[edge.v2 - 1].setGraph((*vGraph)[edge.v1 - 1].getGraph());
            (*vGraph)[edge.v1 - 1].getGraph()->addNode(edge.v2);
         }
      } else {
         if ((*vGraph)[edge.v2 - 1].getGraph()) {
            (*vGraph)[edge.v1 - 1].setGraph((*vGraph)[edge.v2 - 1].getGraph());
            (*vGraph)[edge.v2 - 1].getGraph()->addNode(edge.v1);
         } else {
            graph = new Graph();
            graph->addNode(edge.v1);
            graph->addNode(edge.v2);
            
            (*vGraph)[edge.v1 - 1].setGraph(graph);
            (*vGraph)[edge.v2 - 1].setGraph(graph);
         }
      }
   }
}

void mergeBoundaries(int vertex, vector<GraphNode> *vGraph, const set<int> &prevBoundary) {
   Face *face;
   set<int>::const_iterator sitr;
   Edge candidate;
   Edge best;

   // find vertex on previous boundary that shares an edge with vertex
   for (vector<int>::iterator itr = vertexXFace[vertex - 1].begin(); itr != vertexXFace[vertex - 1].end(); itr++) {
      face = gModel->mFaces[*itr];
      sitr = prevBoundary.find(face->v1);
      if (sitr != prevBoundary.end()) {
         candidate = getEdge(vertex, face->v1);
         if (best.v1 == -1 ||
            (*gFGraph)[best.t1].getSiblings().size() + (*gFGraph)[best.t2].getSiblings().size() < 
            (*gFGraph)[candidate.t1].getSiblings().size() + (*gFGraph)[candidate.t2].getSiblings().size()) {
            best = candidate;
         }
      }
      sitr = prevBoundary.find(face->v2);
      if (sitr != prevBoundary.end()) {
         candidate = getEdge(vertex, face->v2);
         if (best.v1 == -1 ||
            (*gFGraph)[best.t1].getSiblings().size() + (*gFGraph)[best.t2].getSiblings().size() < 
            (*gFGraph)[candidate.t1].getSiblings().size() + (*gFGraph)[candidate.t2].getSiblings().size()) {
            best = candidate;
         }
      }
      sitr = prevBoundary.find(face->v3);
      if (sitr != prevBoundary.end()) {
         candidate = getEdge(vertex, face->v3);
         if (best.v1 == -1 ||
            (*gFGraph)[best.t1].getSiblings().size() + (*gFGraph)[best.t2].getSiblings().size() < 
            (*gFGraph)[candidate.t1].getSiblings().size() + (*gFGraph)[candidate.t2].getSiblings().size()) {
            best = candidate;
         }
      }
   }
   if (best.v1 == -1) {
      cout << "No edge to merge was found." << endl;
      assert(false);
   } else {
      connectBoundaries(vGraph, gFGraph, best);
   }
}

void getRootFace(vector<GraphNode> *fGraph) {
   for (vector<int>::iterator itr = vertexXFace[rootI - 1].begin(); itr != vertexXFace[rootI - 1].end(); itr++) {
      if ((*fGraph)[*itr].getSiblings().size() == 1) {
         rootFace = *itr;
         break;
      }
   }

   if (rootFace == -1) {
      cout << "Root face not found!" << endl;
      assert(false);
   }
}

void convertToTrees() {
   list<int> faces;
   list<int> previousFaces;
   int currentFace;
   int previous = -1;
   int faceCount = 0;

   bool *visited = new bool[gModel->mFaces.size()];
   for (size_t i = 0; i < gModel->mFaces.size(); i++) {
      visited[i] = false;
   }

   faces.push_back(rootFace);
   previousFaces.push_back(-1);

   while (!faces.empty()) {
      currentFace = faces.front();
      faces.pop_front();
      previous = previousFaces.front();
      previousFaces.pop_front();

      (*gFGraph)[currentFace].removeSibling(previous);

      for (vector<int>::const_iterator itr = (*gFGraph)[currentFace].getSiblings().begin();
       itr != (*gFGraph)[currentFace].getSiblings().end(); itr++) {
            if (visited[*itr]) multVisited[*itr] = true;
            visited[*itr] = true;
            faces.push_back(*itr);
            previousFaces.push_back(currentFace);
      }

      faceCount++;
   }

   cout << "Finished: " << faceCount << endl;
   delete visited;
}

void generateTrees(vector<GraphNode> *vGraph, vector<GraphNode> *fGraph) {
   vector<set<int> *> boundaries;
   set<int> *boundary = new set<int>();
   int currentBoundary = 0;
   Face *face;
   Graph *rootVGraph;
   int count = 0;
   int maxItr = 10000;
   int lastMerge = 0;
   int totalBoundarySize = 1;

   set<int>::iterator vitr;
   vector<int>::iterator fitr;

   // insert root vertex as first boundary
   boundary->insert(rootI);
   boundaries.push_back(boundary);

   rootVGraph = new Graph();
   rootVGraph->addNode(rootI);
   (*vGraph)[rootI - 1].setGraph(rootVGraph);

   // while there are vertices and faces to visit
   while(facesVisited < gModel->mFaces.size()) {
      lastMerge = boundaryEdge.size();
      count++;
      boundary = new set<int>();
      // for each vertex in the current boundary
      for (vitr = boundaries[currentBoundary]->begin(); vitr != boundaries[currentBoundary]->end(); vitr++) {
         // for each triangle incident on vertex and not processed
         for (fitr = vertexXFace[*vitr - 1].begin(); fitr != vertexXFace[*vitr - 1].end(); fitr++) {
            if (!faceVisited[*fitr]) {
               faceVisited[*fitr] = true;
               // for each edge on triangle
               face = gModel->mFaces[*fitr];
               processEdge(face->v1, face->v2, *vitr, vGraph, fGraph, boundary);
               processEdge(face->v2, face->v3, *vitr, vGraph, fGraph, boundary);
               processEdge(face->v3, face->v1, *vitr, vGraph, fGraph, boundary);
            }
         }
      }

      totalBoundarySize += boundary->size();
      if (boundary->size() == 0) {
         cout << "Boundary size reached 0" << endl;
         break;
      }

      currentBoundary++;
      boundaries.push_back(boundary);
      for (set<int>::iterator itr = boundaries[currentBoundary]->begin(); itr != boundaries[currentBoundary]->end(); itr++) {
         if ((*vGraph)[rootI - 1].getGraph() != (*vGraph)[*itr - 1].getGraph()) {
            mergeBoundaries(*itr, vGraph, *boundaries[currentBoundary - 1]);
         }
      }

      if (count >= maxItr) break;
   }
   cout << "Faces visited: " << facesVisited << endl;
   cout << "Total Vertices in Boundary: " << totalBoundarySize << endl;

   getRootFace(fGraph);
}

//initialization calls for opengl for static light
//note that we still need to enable lighting in order for it to work
//see keyboard 'l' event
void initializeLighting() {
   //turn on light0
   glEnable(GL_LIGHT0);
   //set up the diffuse, ambient and specular components for the light
   glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diff);
   glLightfv(GL_LIGHT0, GL_AMBIENT, light_amb);
   glLightfv(GL_LIGHT0, GL_SPECULAR, light_spec);
   //specify our lighting model as 1 normal per face
   glShadeModel(GL_FLAT);
   glEnable(GL_LIGHTING);
}

void pos_light() {
   glMatrixMode(GL_MODELVIEW);
   glLightfv(GL_LIGHT0, GL_POSITION, (float *)cLight);
}

void useMaterial(int color) {
   color = color % NUM_MATERIALS;
   glMaterialfv(GL_FRONT, GL_AMBIENT, materials[color].ambient);
   glMaterialfv(GL_FRONT, GL_DIFFUSE, materials[color].diffuse);
   glMaterialfv(GL_FRONT, GL_SPECULAR, materials[color].specular);
   glMaterialfv(GL_FRONT, GL_SHININESS, materials[color].shininess);
}

void setUpView() {
   // this code initializes the viewing transform
   glLoadIdentity();

   // moves viewer along coordinate axes
   gluLookAt(cCamera.x, cCamera.y, cCamera.z, 
             0.0, 0.0, 0.0, 0.0, 1.0, 0.0);

   // move the view back some relative to viewer[] position
   glTranslatef(0.0f,0.0f, 0.0f);

   return;
}

void drawTriangleTree() {
   list<int> faces;
   list<int> previousFaces;
   int currentFace;
   int previous = -1;
   int faceCount = 0;

   faces.push_back(rootFace);
   previousFaces.push_back(-1);

   while (!faces.empty()) {
      currentFace = faces.front();
      faces.pop_front();
      previous = previousFaces.front();
      previousFaces.pop_front();

      drawFace(gModel->mFaces[currentFace], (*gFGraph)[currentFace].getSiblings().size(), currentFace);
      for (vector<int>::const_iterator itr = (*gFGraph)[currentFace].getSiblings().begin();
       itr != (*gFGraph)[currentFace].getSiblings().end(); itr++) {
         if (*itr != previous) {
            faces.push_back(*itr);
            previousFaces.push_back(currentFace);
         }
      }
      faceCount++;
   }

   
   /*
   drawFace(gModel->mFaces[face], (*gFGraph)[face].getSiblings().size());

   for (vector<int>::const_iterator itr = (*gFGraph)[face].getSiblings().begin(); itr != (*gFGraph)[face].getSiblings().end(); itr++) {
      if (*itr != previous) {
         drawTriangleTree(*itr, face);
      }
   }
   */
}

void drawVertexTree(int vertex) {
   list<int> vertices;
   list<int> previousVertices;
   int currentVertex;
   int previous = -1;
   int vertexCount = 0;

   vertices.push_back(rootI);
   previousVertices.push_back(rootI);

   while (!vertices.empty()) {
      currentVertex = vertices.front();
      vertices.pop_front();
      previous = previousVertices.front();
      previousVertices.pop_front();

      for (vector<int>::const_iterator itr = (*gVGraph)[currentVertex - 1].getSiblings().begin();
       itr != (*gVGraph)[currentVertex - 1].getSiblings().end(); itr++) {
         if (*itr != previous) {
            vertices.push_back(*itr);
            previousVertices.push_back(currentVertex);
            drawEdge(Edge(currentVertex, *itr), false);
         }
      }
      vertexCount++;
   }
}

//--------------------------------------------------------
//  Set up the objects

void drawMesh()
{
   list<int>::iterator iitr;
   vector<set<int> *>::iterator vitr;
   set<int>::iterator sitr;
   vector<Edge>::iterator eitr;
   int color = 0;
   int face = 0;

   glPushMatrix();
   useMaterial(color);
   pos_light();
   glScalef(1.0 /(float)gModel->max_extent, 1.0 /(float)gModel->max_extent,
    1.0 /(float)gModel->max_extent);

   glTranslatef(-(gModel->mCenter.x), -(gModel->mCenter.y), -(gModel->mCenter.z));

   if (drawMode == kModel) {
      for(size_t j = 0; j < gModel->mFaces.size(); j++) {
         drawFace(gModel->mFaces[j], -1, j);
      }
   } else if (drawMode == kGraph) {
      for (vector<GraphNode>::iterator nitr = gFGraph->begin(); nitr != gFGraph->end(); nitr++, face++) {
         drawFace(gModel->mFaces[face], (*nitr).getSiblings().size(), face);
      }
      
      for(size_t j = 0; j < gModel->mVertices.size(); j++) {
         for (vector<int>::const_iterator titr = (*gVGraph)[j].getSiblings().begin(); titr != (*gVGraph)[j].getSiblings().end(); titr++) {
            drawEdge(Edge(j + 1, *titr), false);
         }
      }
      
      for(size_t j = 0; j < gModel->mVertices.size(); j++) {
         for (eitr = (*gModel->mEdges)[j].begin(); eitr != (*gModel->mEdges)[j].end(); eitr++) {
            drawEdge(*eitr, true);
         }
      }
   } else if (drawMode == kTree) {
      drawTriangleTree();
      drawVertexTree(rootI);

      for(size_t j = 0; j < gModel->mVertices.size(); j++) {
         for (eitr = (*gModel->mEdges)[j].begin(); eitr != (*gModel->mEdges)[j].end(); eitr++) {
            drawEdge(*eitr, true);
         }
      }
   }
   drawRootNormal();
   
   glPopMatrix();
}

void drawRootNormal(){
   glDisable(GL_LIGHTING);
   glLineWidth(2.0);
   glPointSize(1.0);
   glBegin(GL_LINES);
   glColor3f(1.0, 0.0, 0.0);
   glVertex3f(rootV.x, rootV.y, rootV.z);
   glVertex3f(rootV.x + ROOT_NORM_SCALE * gModel->mNormals[rootI-1]->x / BIG_NUM, 
    rootV.y + ROOT_NORM_SCALE * gModel->mNormals[rootI-1]->y / BIG_NUM,
    rootV.z + ROOT_NORM_SCALE * gModel->mNormals[rootI-1]->z / BIG_NUM);
   glEnd();
   glEnable(GL_LIGHTING);
}

void drawEdge(const Edge &e, bool isMarchingEdge) {
   glDisable(GL_LIGHTING);
   if (isMarchingEdge)
      glLineWidth(1.0);
   else
      glLineWidth(2.0);
   glBegin(GL_LINES);

   if (isMarchingEdge) {
      glColor3f(0.3, 0.3, 0.3);
   } else {
      glColor3f(1.0, 1.0, 1.0);
   }
   glVertex3f(gModel->mVertices[e.v1-1]->x + gModel->mNormals[e.v1-1]->x / BIG_NUM, 
    gModel->mVertices[e.v1-1]->y + gModel->mNormals[e.v1-1]->y  / BIG_NUM,
    gModel->mVertices[e.v1 - 1]->z + gModel->mNormals[e.v1-1]->z  / BIG_NUM);
   glVertex3f(gModel->mVertices[e.v2 - 1]->x + gModel->mNormals[e.v2-1]->x  / BIG_NUM, 
    gModel->mVertices[e.v2 - 1]->y + gModel->mNormals[e.v2-1]->y  / BIG_NUM,
    gModel->mVertices[e.v2 - 1]->z + gModel->mNormals[e.v2-1]->z  / BIG_NUM);
   glEnd();
   glEnable(GL_LIGHTING);
}

void drawFace(Face *t, int size, int faceNum)
{
   MyVector fcenter;

   glNormal3f(t->mNormal.x / gModel->max_extent,
    t->mNormal.y / gModel->max_extent,
    t->mNormal.z / gModel->max_extent);

   if (drawMode != kModel) glDisable(GL_LIGHTING);
   glBegin(GL_POLYGON);

   if (drawMode != kModel) {
      if (debugMode && multVisited[faceNum]) {
         glColor3f(1.0, 0.0, 0.0);
      } else {
         size++;
         if (gModel->mFaces[rootFace] == t)
            glColor3f(0.0, 0.0, 1.0);
         else {
            if (size == 0) { // bright red
               glColor3f(1.0, 0.0, 0.0);
            } else if (size == 1) { // green
               glColor3f(0.0, 0.6, 0.0);
            } else if (size == 2) { // yellow
               glColor3f(0.7, 0.7, 0.0);
            } else if (size == 3) { // brown
               glColor3f(0.4, 0.2, 0.2);
            } else { // white
               glColor3f(1.0, 1.0, 1.0);
            }
         }
      }
   }

   glVertex3f(gModel->mVertices[t->v1 - 1]->x, 
   gModel->mVertices[t->v1 - 1]->y,
   gModel->mVertices[t->v1 - 1]->z);

   glVertex3f(gModel->mVertices[t->v2 - 1]->x, 
   gModel->mVertices[t->v2 - 1]->y,
   gModel->mVertices[t->v2 - 1]->z);

   glVertex3f(gModel->mVertices[t->v3 - 1]->x, 
   gModel->mVertices[t->v3 - 1]->y,
   gModel->mVertices[t->v3 - 1]->z);

   glEnd();

   if (drawMode != kModel) glEnable(GL_LIGHTING);
}

//-----------------------------------------------------------
//  Callback functions

void reshapeCallback(int w, int h) {
   GW = w;
   GH = h;

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();

   gluPerspective(50, (float)GW/GH, 0.1, 15);
   glMatrixMode(GL_MODELVIEW);
   glViewport(0, 0, w, h);

   glutPostRedisplay();
}

void calculateCamera()
{
   cCamera.x = cRadius * cos(cPhi) * cos(3.14159 / 2 - cTheta);
   cCamera.y = cRadius * sin(cPhi);
   cCamera.z = cRadius * cos(cPhi) * cos(cTheta);
}

void processSpecialKeys(int key, int x, int y) {
   switch( key ){
   case GLUT_KEY_DOWN:
	   cPhi += INCREMENT;
      if (cPhi > MAX_PHI) cPhi = MAX_PHI;
	   calculateCamera();
		glutPostRedisplay();
      break;
   case GLUT_KEY_UP:
		phi = cPhi -= INCREMENT;
      if (cPhi < MIN_PHI) cPhi = MIN_PHI;
		calculateCamera();
		glutPostRedisplay();
      break;
   case GLUT_KEY_LEFT:
      theta = cTheta += INCREMENT;
      calculateCamera();
      glutPostRedisplay();
      break;
   case GLUT_KEY_RIGHT:
      theta = cTheta -= INCREMENT;
      calculateCamera();
      glutPostRedisplay();
      break;
   }
}

//---------------------------------------------------------
//  Main routines

void display (void) {
   // this code executes whenever the window is redrawn (when opened,
   //   moved, resized, etc.
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   // set the viewing transform
   setUpView();

   // start drawing objects
   drawMesh();

   glutSwapBuffers();
}

void processNormalKeys(unsigned char key, int x, int y )
{
   switch(key) {
      case 'z':
         cRadius -= ZINCREMENT;
         if (cRadius < MIN_RAD) cRadius = MIN_RAD;
         calculateCamera();
         glutPostRedisplay();
         break;
      case 'x':
         cRadius += ZINCREMENT;
         if (cRadius > MAX_RAD) cRadius = MAX_RAD;
         calculateCamera();
         glutPostRedisplay();
         break;
      case 't':
         drawMode = kTree;
         cout << "Rendering based on trees." << endl;
         glutPostRedisplay();
         break;
      case 'm':
         drawMode = kModel;
         cout << "Rendering based on model." << endl;
         glutPostRedisplay();
         break;
      case 'g':
         drawMode = kGraph;
         cout << "Rendering based on graph." << endl;
         glutPostRedisplay();
         break;
      case 'd':
         debugMode = !debugMode;
         glutPostRedisplay();
         break;
      default:
         break;
   }
}

void initializeCamera(){
   cPhi = PI / 8.0;
   cTheta = -PI / 4.0;
   cRadius = 3.0;
}

void selectRootVertex(){
   srand(time(0));
   rootI = gModel->mVertices.size() * (rand() % 1001) / 1000.0;
   rootV = *(gModel->mVertices[rootI - 1]);
}

void runSanityChecks() {
   list<int> vertices;
   list<int> previousVertices;
   int currentVertex;
   int previous = -1;
   int vertexCount = 0;
   int verticesInGraph = 0;
   int differentGraph = 0;

   vertices.push_back(rootI);
   previousVertices.push_back(rootI);

   cout << "----Checking max face connectivity" << endl;
   for (size_t i = 0; i < gModel->mFaces.size(); i++) {
      if ((*gFGraph)[i].getSiblings().size() > 3) {
         cout << "connections: " << (*gFGraph)[i].getSiblings().size() << endl;
         assert(false);
      }
   }

   cout << "----Checking vertex count in graph" << endl;
   while (!vertices.empty()) {
      verticesInGraph++;
      currentVertex = vertices.front();
      vertices.pop_front();
      previous = previousVertices.front();
      previousVertices.pop_front();

      for (vector<int>::const_iterator itr = (*gVGraph)[currentVertex - 1].getSiblings().begin();
       itr != (*gVGraph)[currentVertex - 1].getSiblings().end(); itr++) {
         if (*itr != previous) {
            vertices.push_back(*itr);
            previousVertices.push_back(currentVertex);
         }
      }
   }
   if (verticesInGraph != gModel->mVertices.size()) {
      cout << "Number of vertices in graph does not match model: " << verticesInGraph << " / " << gModel->mVertices.size() << endl;
   }
   bool found;
   cout << "----Checking each vertices graph" << endl;
   for (size_t i = 0; i < gModel->mVertices.size(); i++) {
      found = false;
      if (!(*gVGraph)[i].getGraph()) {
         vertexCount++;
         continue;
      }
      if ((*gVGraph)[i].getGraph() != (*gVGraph)[rootI - 1].getGraph()) {
         differentGraph++;
      }
      for (vector<int>::const_iterator itr = (*gVGraph)[i].getGraph()->getNodes().begin(); itr != (*gVGraph)[i].getGraph()->getNodes().end(); itr++) {
         if (*itr - 1 == i) {
            found = true;
            break;
         }
      }
      if (!found) {
         cout << "Vertex not in its graph!" << endl;
         assert(false);
      }
   }
   if (differentGraph > 0) {
      cout << "Vertices with different graphs than root vertex: " << differentGraph << endl;
   } else {
      cout << "All vertices share the same graph." << endl;
   }
   if (vertexCount > 0) {
      cout << "Vertices with no graph: " << vertexCount << endl;
   } else {
      cout << "Every vertex has a graph." << endl;
   }
   cout << "----Ensuring face graph does not cross vertex graph" << endl;
   for (size_t i = 0; i < (*gModel->mEdges).size(); i++) {
      for (size_t j = 0; j < (*gModel->mEdges)[i].size(); j++) {
         if ((*gVGraph)[((*gModel->mEdges)[i][j].v1 - 1)].contains((*gModel->mEdges)[i][j].v2) &&
            (*gFGraph)[(*gModel->mEdges)[i][j].t1].contains((*gModel->mEdges)[i][j].t2)) {
            cout << "vertices and faces connected accross one edge!!!" << endl;
            assert(false);
         }
      }
   }
}

// create a double buffered 500x500 pixel color window
int main(int argc, char* argv[]) {
   GW = 600;
   GH = 600;

   if (argc == 2) {
      gModel = new Model(argv[1]);
   } else {
      gModel = new Model("dragon10k.m");
   }

   if (!gModel) {
      cout << "No model loaded, exiting." << endl;
      exit(-1);
   }

   cout << "Faces read in: " << gModel->mFaces.size() << endl;
   cout << "Vertices read in: " << gModel->mVertices.size() << endl;

   gVGraph = new vector<GraphNode>(gModel->mVertices.size());
   gFGraph = new vector<GraphNode>(gModel->mFaces.size());
   faceVisited = new bool[gModel->mFaces.size()];
   multVisited = new bool[gModel->mFaces.size()];
   for (size_t i = 0; i < gModel->mFaces.size(); i++) {
      faceVisited[i] = false;
      multVisited[i] = false;
   }

   selectRootVertex();
   cout << "--Generating edges" << endl;
   generateEdges(gModel);
   cout << "--Generating graphs" << endl;
   generateTrees(gVGraph, gFGraph);
   cout << "*********************" << endl;

   if (debugMode) {
      cout << "--Running sanity checks" << endl;
      runSanityChecks();
   }

   cout << "--Converting graphs to trees" << endl;
   convertToTrees();

   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
   glutInitWindowSize(1000, 1000);
   glutInitWindowPosition(100, 100);
   glutCreateWindow("Topological Surgery");
   glutSpecialFunc(processSpecialKeys);
   glutKeyboardFunc(processNormalKeys);

	glEnable(GL_DEPTH_TEST);
   initializeCamera();
   calculateCamera();
   initializeLighting();
	glutDisplayFunc(display);
   glutReshapeFunc(reshapeCallback);

	glutMainLoop();
	return 0;
}

//---------------------------------------------------------
//  Utility functions

void normalize(float v[3]) {
   // normalize v[] and return the result in v[]
   // from OpenGL Programming Guide, p. 58
   GLfloat d = sqrt(v[1]*v[1] + v[2]*v[2] + v[3]*v[3]);
   if (d == 0.0) {
      printf("zero length vector");
      return;
   }
   v[1] = v[1]/d; v[2] = v[2]/d; v[3] = v[3]/d;
}

void normCrossProd(float v1[3], float v2[3], float out[3]) {
   // cross v1[] and v2[] and return the result in out[]
   // from OpenGL Programming Guide, p. 58
   out[0] = v1[1]*v2[2] - v1[2]*v2[1];
   out[1] = v1[2]*v2[0] - v2[0]*v2[2];
   out[2] = v1[0]*v2[1] - v1[1]*v2[0];
   normalize(out);
}

